//Manuscript Title: Tomographic diffracitive micrscopy and multiview profilometry
//                  with flexible aberration correction
//
//Authors: H. Liu, J. Bailleul, B. Simon, M. Debailleul, B. Colicchio, and O. Haeberlé
//Program title: CorrectionAberration
//Distribution format: Zip
//Download link: to come
//Licensing provisions: CeCILL-B
//Journal reference: Applied optics, Vol. 53, issue 4, p748-755 (2014)
//Programming language: C++
//Computer: General computers
//Operating Synstem: All
//Keywords: Aberration compensation; Holography; Tomographic imaging; Fourier optics and signal processing;
//External routines: Opencv
//
//Nature of problem: Automatic aberration compensation in holographic imaging
//Solution method: Polynomial approximation of the aberration function,
//                 by using a random point selection in the background
//Running time: 0,3s with an Intel quadcore i7 processor and 16 Gbytes of memory

#include <iostream>
#include <fstream>
#include <sstream>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <cxcore.h>
#include <highgui.h>
#include <cv.h>
using namespace cv;
using namespace std;

/// Function header
void threshCallback(int, void*);
double poly2DEval(Mat coefficients, int deg,int x, int y);
int sizePoly2D(int deg);
int countM(Mat mask);
void compuBackgr(Mat coefficients, int deg, Mat imageBackgr);
void compuPoly(Mat imagebrut, Mat mask, Mat& polynomial, int deg, bool method);
void aberCorr(Mat image, Mat mask);

Mat src;
Mat src_gray;
int thresh = 100;
int max_thresh = 255;
RNG rng(12345);

int main( int argc, char** argv )
{
    /// Load source image
    src = imread("CD.jpg", 0); /// 0 gray image, 1 color image

    /// Convert image to gray, if color image
    //cvtColor(src, src_gray, CV_BGR2GRAY);

    /// OpenCV Gaussian blur function
    GaussianBlur(src, src_gray, Size( 5, 5 ), 2, 2);

    /// Create Window
    char* source_window = "Source";
    namedWindow(source_window, CV_WINDOW_AUTOSIZE);
    imshow(source_window, src);

    createTrackbar("Canny thresh:", "Source", &thresh, max_thresh, threshCallback);
    threshCallback(0, 0);
    waitKey(0);

    /// Image_mask.jpg is computed by threshCallback()
    Mat mask = imread("Image_mask.jpg", 0);

    /// Aberration correction
    aberCorr(src, mask);

    waitKey(0);
    return 0;
}

/// function threshCallback
void threshCallback(int, void*)
{
    Mat canny_output;
    Mat src_filtering;
    vector<vector<Point> > contours;
    vector<Vec4i> hierarchy;

    /// Detect edges using canny
    Canny(src_gray, canny_output, thresh, thresh*2, 3);
    /// Detect edges using Threshold
    threshold(src_gray, canny_output, thresh, 255, THRESH_BINARY);

    /// Dilate helps to remove potential holes between edge segments
    dilate(canny_output, canny_output, Mat(), Point(-1,-1));

    /// Find contours
    findContours(canny_output, contours, hierarchy, CV_RETR_TREE, CV_CHAIN_APPROX_SIMPLE, Point(0, 0));

    /// Find the convex hull object for each contour
    vector<vector<Point> >hull(contours.size());
    for (int i = 0; i < contours.size(); i ++)
    {
        convexHull(Mat(contours[i]), hull[i], false);
    }

    /// Draw contours
    Mat drawing = Mat::zeros(canny_output.size(), CV_8UC3);
    Mat drawing_gray;
    for (int i = 0; i < contours.size(); i ++)
    {
        Scalar color = Scalar(rng.uniform(0, 255), rng.uniform(0,255), rng.uniform(0,255));
        /// hull results
        drawContours(drawing, hull, i, Scalar(255,255,255), CV_FILLED, 8, vector<Vec4i>(), 0, Point());
    }

    /// Inverse gray and convert rgb to gray
    convertScaleAbs(drawing, drawing, -1, 255);
    cvtColor(drawing, drawing_gray, CV_BGR2GRAY);

    /// Show in a window
    namedWindow("Image_mask", CV_WINDOW_AUTOSIZE);
    imshow("Image_mask", drawing_gray);
    imwrite("Image_mask.jpg",drawing_gray);
}

/// Compute the sum of the polynomial, for deg 3: 1+x+x²+x³+y+xy+x²y+y²+xy²+y³
double poly2DEval(Mat coefficients, int deg, int x, int y)
{
    int k = 0;
    double sum = 0;
    for (int i = 0; i <= deg; i ++)
    {
        int j = 0;
        while ((i+j) <= deg)
        {
            sum += (coefficients.at<double>(k)) * pow(x,i) * pow(y,j);
            k ++;
            j ++;
        }
    }
    return sum;
}

/// Return the size of polynomial
int sizePoly2D(int deg)
{
    int j = 0, size = 0;
    for (int i = 0; i <= deg; i ++)
    {
        while ((i+j) <= deg)
        {
            size ++;
            j ++;
        }
        j = 0;
    }
    return size;
}

/// Count pixels in mask
int countM(Mat mask)
{
    int count = 0;
    for (int i = 0; i < mask.rows; i ++)
    {
        for (int j = 0; j < mask.cols; j ++)
        {
            if (mask.at<uchar>(j,i) > 10)
                count ++;
        }
    }
    return count;
}

/// Compute the coef of polynomial (Least Squares method)
void compuPoly(Mat imagebrut, Mat mask, Mat& polynomial, int deg, bool method)
{
    int nbCols = sizePoly2D(deg);
    //int nbRows = countNonZero(mask); /// Opencv function
    int nbRows = countM(mask);
    int col, k = 0;
    Mat B(Size(nbCols,nbRows), CV_64F);
    Mat Bt(Size(nbRows,nbCols), CV_64F);
    Mat f(nbRows, 1, CV_64F);

    for (int y = 0; y < imagebrut.rows; y ++)
    {
        for (int x = 0; x < imagebrut.cols; x ++)
        {
            col = 0;
            if (mask.at<uchar>(y,x) > 10)
            {
                f.at<double>(k) = (double)imagebrut.at<uchar>(y,x); /// uchar--> double for .bin files
                for (int i = 0; i <= deg; i ++)
                {
                    int j = 0;
                    while((i+j) <= deg)
                    {
                        B.at<double>(k,col) = pow((double)x,i) * pow((double)y,j);
                        col ++;
                        j ++;
                    }
                }
                k ++;
            }
        }
    }

    Mat coef(nbCols, 1, CV_64F);
    Mat D(Size(nbCols, nbRows), CV_64F);
    Mat invD(Size(nbCols, nbRows), CV_64F);
    /// Use OpenCV solve() function to solve the linear system
    if (method)  cv::solve(B, f, coef, DECOMP_NORMAL);
    else
    {
        /// Use OpenCV matrix inversion operators to solve the linear system
        cv::transpose(B, Bt);
        D = Bt * B;
        cv::invert(D, invD);
        coef = (invD * Bt) *f;
    }
    coef.copyTo(polynomial);
}

/// Compute the background image with the coef of polynomial
void compuBackgr(Mat coefficients, int deg, Mat imageBackgr)
{
    for (int y = 0; y < imageBackgr.rows; y ++)
    {
        for (int x = 0; x < imageBackgr.cols; x ++)
        {
            imageBackgr.at<double>(y,x) = poly2DEval(coefficients, deg, x, y);
        }
    }
}

/// Apply the aberration correction
void aberCorr(Mat image, Mat mask)
{
    int degpoly = 3;
    Mat coefsolve;
    /// Compute the coef of polynomial (Least Squares method)
    compuPoly(image, mask, coefsolve, degpoly, true);
    //cout << coefsolve << endl;
    Mat resultatpoly(image.rows, image.cols, CV_64F);
    /// Compute the background image with the coef of polynomial
    compuBackgr(coefsolve, degpoly, resultatpoly);
    imwrite("Image_background.jpg", resultatpoly);
    Mat tmp;
    image.convertTo(tmp, CV_64F);
    /// Aberration substraction
    tmp = tmp - resultatpoly;
    /// The result is stored in Image_fit.jpg
    imwrite("Image_fit.jpg", tmp);
}
